import flask, os, os.path, time, requests, websockets, json, asyncio, urllib
import multiprocessing, base64, functools, inspect, datetime
from cryptography.hazmat.primitives import serialization, hashes, hmac
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from itsdangerous import BadSignature

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from utils import project_path, RouteLobby, secret_key
from tokens import refresh_access, default_timeouts, authorized
from group import AccessGroupRef
from store import ThreadDB, FKDatabase
from utils import relpath, dict_names
from builder import LoginBuilder, app as DefaultBuilder
from interface import OAuthBlueprint

end_locals()

# custom made protocol; approach with skepticism
class Handshake:
    """
    threat model: lateral movement on a shared machine
    assumptions:
      - the WS URI's numeric port can be hijecked by a malicious process
      - anything in the run directory is privledged including unix sockets
      - run has a secret_key needed to verify the flask session's login cookie
      - HTTPS certs validate primary's authority over secondary

    requirements:
      - secondaries shouldn't be able to immitate the primary
      - significantly delayed/successfully replayed messages are failures
      - the primary should also be able to connect via port number
    """

    signing_params = (padding.PSS(
        mgf=padding.MGF1(hashes.SHA256()),
        salt_length=padding.PSS.MAX_LENGTH),
        hashes.SHA256())
    otp_timeout_ms = 5000

    def __init__(self, root_path=None):
        self._root_path = root_path or project_path()
        self.server_unix_path = self.run_path("server.sock")
        self.client_unix_path = self.run_path("client.sock")

    def root_path(self, *a):
        return os.path.join(self._root_path, *a)

    def run_path(self, *a):
        return self.root_path("run", *a)

    @staticmethod
    def serialized_public(key):
        return key.public_key().public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo)

    # guarantees self.key
    def keypair(self):
        if os.path.exists(self.run_path("private.pem")):
            with open(self.run_path("private.pem"), "rb") as fp:
                key = serialization.load_pem_private_key(
                    fp.read(), password=None)
        else:
            key = rsa.generate_private_key(
                public_exponent=65537,
                key_size=2048
            )

            with open(self.run_path("private.pem"), "wb") as fp:
                fp.write(key.private_bytes(
                    serialization.Encoding.PEM,
                    serialization.PrivateFormat.PKCS8,
                    serialization.NoEncryption()))

            with open(self.run_path("public.pem"), "wb") as fp:
                fp.write(self.serialized_public(key))

        self.key = key

    # returns exit status as bool
    public = None
    def load_public(self):
        if os.path.exists(self.run_path("public.pem")):
            with open(self.run_path("public.pem"), 'rb') as fp:
                self.public = serialization.load_pem_public_key(fp.read())
            return False
        return True

    @staticmethod
    def now():
        return int(time.time() * 10 ** 3).to_bytes(8, 'big')

    # limits app secret key to 128, 192, or 256 bits
    @property
    def secret(self):
        return secret_key(self.run_path())

    _salt = None
    @property
    def salt(self):
        if self._salt is None:
            self._salt = os.urandom(32)
        return self._salt

    # hmac is 0x20 long, private key signature is 0x100 long
    # client: proves possession of the shared_secret as OTP
    #    0:   8 is timestamp
    #    8:0x28 is timestamp signed by shared_secret
    # 0x28:0x48 is timestamp plus salt signed by shared_secret
    def client_syn(self):
        message = self.now()
        h = hmac.HMAC(self.secret, hashes.SHA256())
        h.update(message)
        g = h.copy()
        g.update(self.salt)
        return message + h.finalize() + g.finalize()

    def otp(self):
        message = self.now()
        h = hmac.HMAC(self.secret, hashes.SHA256())
        h.update(message)
        return message + h.finalize()

    def timestamp_expiration(self, data):
        latency = time.time() - int.from_bytes(data[:8], 'big')
        assert latency < self.otp_timeout_ms, "signature timeout"

    # verifies the OTP timing but not salt
    def server_timer(self, data):
        self.timestamp_expiration(data)
        h = hmac.HMAC(self.secret, hashes.SHA256())
        h.update(data[:8])
        g = h.copy()
        h.verify(data[8:0x28])
        return g

    # verifies OTP timing and salt
    def client_timer(self, data):
        h = server_timer(data)
        g = h.copy()
        h.update(self.salt)
        h.verify(data[0x28:0x48])
        return g

    # saves/returns the AES encrypted public key
    def save_public(self, serialized):
        with open(self.run_path("public.pem"), "wb") as fp:
            fp.write(serialized)
        public = serialization.load_pem_public_key(serialized)
        return public

    # server: verifies the client has the shared secret
    #         proves possession of the shared secret
    #         proves possession of the private key
    #         sends private key encrypted with the shared secret
    #     0: 0x48 is OTP
    #  0x48: 0x68 is salted hash signed by shared secret
    #  0x68:0x168 is salted hash signed by private key
    # 0x168:      is AES encrypted public key with above as nonce
    def server_sign_otp(self, otp):
        g = self.server_timer(otp)
        g.update(data[0x28:0x48])
        return otp + g.finalize() + \
            self.key.sign(self.secret, otp[0x28:0x48], *self.signing_params) + \
            AESGCM(self.secret).encrypt(
                otp[0x28:0x34], self.serialized_public(self.key.public_key()),
                None)

    # client: verifies the OTP is still valid
    #         verifies the OTP was generated with the shared secret
    #         verifies the server has the shared secret
    #         decrypts and stores the public key
    #         verifies the OTP was signed with the corresponding private key
    def client_verify_ack(self, reply):
        h = self.client_timer(reply)
        h.update(data[0x28:0x48])
        h.verify(data[0x48:0x68])
        public = self.save_public(AESGCM(self.secret).decrypt(
            reply[0x28:0x34], reply[0x168:], None))
        public.verify(reply[0x68:0x168], reply[0x28:0x48], *self.signing_params)
        self.public = public
        return public

    def server_send(self, data):
        nonce = os.urandom(12)
        message = nonce + AESGCM(self.secret).encrypt(
            nonce, self.now() + data.encode(), None)
        signature = self.key.sign(message, *self.signing_params)
        return signature + message

    def client_recieve(self, data):
        self.public.verify(data[:0x100], data[0x100:], *self.signing_params)
        data = AESGCM(self.secret).decrypt(
            data[0x100:0x10C], data[0x10C:], None)
        self.timestamp_expiration(data)
        return data[8:].decode()

    def client_send(self, data):
        nonce = os.urandom(12)
        return nonce + AESGCM(self.secret).encrypt(
            nonce, self.now() + data.encode(), None)

    def server_recieve(self, data):
        data = AESGCM(self.secret).decrypt(data[:12], data[12:], None)
        self.timestamp_expiration(data)
        return data[8:].decode()

    def server_sync(self, data):
        message = self.now()
        signature = self.key.sign(message + data.encode(), *self.signing_params)
        return signature + message + data.encode()

    def server_verify(self, data):
        self.key.public_key().verify(
            data[:0x100], data[0x100:], *self.signing_params)
        return self.server_send(data[0x108:])

    def _fork(self, ws):
        e = multiprocessing.Event()
        multiprocessing.Process(target=ws.run, args=(e,), daemon=True).start()
        e.wait()
        return ws

server_lobby = RouteLobby()

class ServerBP(Handshake):
    def __init__(
            self, db=None, host="localhost", port=8001, cache=None,
            lease_timeout=default_timeouts[0],
            refresh_timeout=default_timeouts[1], remote=None, root_path=None):
        super().__init__(root_path)
        self.a = (host, port, cache, lease_timeout, refresh_timeout)
        self.db, self.timeouts = db, self.a[-2:]
        self.bp = flask.Blueprint("ws_handshake", __name__, url_prefix="/ws")
        self.sock = (lambda: websockets.connect(remote)) if remote else \
            (lambda: websockets.unix_connect(self.server_unix_path))
        self.encoding = self.server_sync if remote else lambda x: x
        if db is not None:
            server_lobby.register_lobby(self.bp, self)
        self.keypair()

    def _fork(self):
        return super()._fork(
            ServerWS(*self.a, root_path=self.root_path()))

    @server_lobby.route("/syn", methods=["POST"])
    def syn(self):
        try:
            return self.server_sign_otp(flask.request.data)
        except:
            flask.abort(401)

    @server_lobby.route("/updates", methods=["POST"])
    def updates(self):
        try:
            self.server_timer(flask.request.data)
        except:
            flask.abort(401)
        since = flask.request.args.get("since", 0)
        return json.dumps({
            "action": "deauthorized",
            "data": self.db().queryall(
                "SELECT rowid, revoked_time, refresh, refresh_time "
                "FROM revoked WHERE revoked_time>?", (since,)),
            "timeouts": self.timeouts,
        })

    async def send_deauthorize(self, row, revoked_time, refresh, refresh_time):
        async with self.sock() as websocket:
            await websocket.send(self.encoding(json.dumps({
                "action": "deauthorize",
                "data": [[row, revoked_time, refresh, refresh_time]]})))

    def deauthorize(self, *a, **kw):
        try:
            asyncio.run(self.send_deauthorize(*a, **kw))
        except ConnectionRefusedError:
            pass

class WSHandshake(Handshake):
    _db = None
    def db(self):
        if self._db is None:
            self._db = ThreadDB(
                self.run_path("users.db"), relpath("schema.sql"))
        return self._db

class FunctionList(dict):
    def __call__(self, f):
        self[f.__name__] = f
        return f

actionable = FunctionList()

class SessionLoader(flask.sessions.SecureCookieSessionInterface, Handshake):
    permanent_session_lifetime = datetime.timedelta(days=31)

    def get_signing_serializer(self, app=None):
        return super().get_signing_serializer(
            type("key_wrapper", (), {"secret_key": self.secret})()
            if app is None else app)

    def open(self, val):
        s = self.get_signing_serializer()
        if s is None:
            return None
        if not val:
            return self.session_class()
        max_age = int(self.permanent_session_lifetime.total_seconds())
        try:
            data = s.loads(val, max_age=max_age)
            return self.session_class(data)
        except BadSignature:
            return self.session_class()

    def save(self, session):
        return self.get_signing_serializer().dumps(dict(session))

class ServerWS(WSHandshake):
    def __init__(
            self, host="localhost", port=8001, cache=None,
            lease_timeout=default_timeouts[0],
            refresh_timeout=default_timeouts[1], *, root_path=None):
        super().__init__(root_path)
        self.host, self.port = host, port
        self.cache, self.secondaries = cache, set()
        self.lease_timeout = lease_timeout
        self.refresh_timeout = refresh_timeout
        self.session_loader = SessionLoader(root_path)

    _key = None
    @property
    def key(self):
        self.keypair()
        return self._key

    @key.setter
    def key(self, value):
        self._key = value

    _app = None
    def app(self):
        # can no longer coexist with the server module
        if self._app is None:
            self._app = DefaultBuilder
            bp = OAuthBlueprint(root_path=self.root_path())
            bp.session = staticmethod(lambda app: flask.g.session)
            self._app.register_blueprint(bp)
        return self._app

    @actionable
    def refresh(self, auth):
        db = self.db().begin()
        cached = self.cache and self.cache.get(auth)
        if cached is None:
            timing = db.queryone(
                "SELECT authtime, refresh_time FROM active WHERE refresh=?",
                (auth,))
            if timing is None:
                return json.dumps(None)
            authtime, refresh_time = timing
        else:
            _, _, authtime, refresh_time = cached

        if self.refresh_timeout and int(
                time.time()) - authtime > self.refresh_timeout:
            if cached is not None:
                self.cache.delete(auth)
            return json.dumps(None)

        updated, write, refresh_time = refresh_access(
            db, auth, refresh_time, self.lease_timeout,
            cached is not None)
        # TODO: cache on updated

        if write:
            db.commit()
        db.close()

        return refresh_time

    @actionable
    def access_query(self, user, access_group, sep='/'):
        return dict_names(AccessGroupRef.reconstruct(
            self.db, access_group, sep).vet(user))

    @actionable
    def ensure_access(self, access_group, sep='/', owner=None):
        group = AccessGroupRef.reconstruct(self.db, access_group, sep, owner)
        group.ensure()
        return group.uuid

    @actionable
    def add_access(self, user, access_group, sep='/'):
        return AccessGroupRef.reconstruct(
            self.db, access_group, sep).add_user(user)

    @actionable
    def remove_access(self, guild):
        return AccessGroupRef.reconstruct(
            self.db, access_id=guild).remove_access(user)

    @actionable
    def from_value(self, cookie, remote_addr, group=None, sep='/'):
        session = self.session_loader.open(cookie)
        with self.app().app_context():
            # the flask dance blueprints modify the current context
            # with before_app_request for all requests to allow lookup
            for ctx_setup in flask.current_app.before_request_funcs[None]:
                ctx_setup()
            flask.g.session = session
            with self.app().test_request_context(
                    environ_base={'REMOTE_ADDR': remote_addr}):
                if not authorized(session):
                    return {"user": None, "cookie": None, "modified": True}
        user = {
                "id": session["user"],
                "name": session["name"],
                "picture": session["picture"],
            }
        if group is not None:
            user["via"] = {"access": self.access_query(user["id"], group, sep)}
        return {
            "user": user, "modified": session.modified,
            "cookie": self.session_loader.save(session)
        }

    async def local_primary(self, ws, init):
        websockets.broadcast(self.secondaries, self.server_send(init))
        async for message in ws:
            websockets.broadcast(self.secondaries, self.server_send(message))

    async def local_router(self, ws):
        message = await ws.recv()
        data = json.loads(message)
        if data["action"] == "deauthorize":
            await self.local_primary(ws, message)
        else:
            await ws.send(self.handler(message))
            async for message in ws:
                await ws.send(self.handler(message))

    def relay(self, message):
        websockets.broadcast(self.secondaries, self.server_verify(message))

    async def remote_primary(self, ws, init):
        self.relay(init)
        async for message in ws:
            self.relay(message)

    def handler(self, message):
        data = json.loads(message)
        action = data.pop("action")
        assert action in actionable
        return json.dumps(actionable[action](self, **data))

    async def secondary(self, ws, init):
        self.server_timer(base64.b64decode(init["data"]))
        subscribed = init["action"] == "subscribe"
        if subscribed:
            self.secondaries.add(ws)
        try:
            async for message in ws:
                refresh_token = self.handler(self.server_recieve(message))
                await ws.send(self.server_send(refresh_token))
        finally:
            if subscribed:
                self.secondaries.remove(ws)

    async def remote_router(self, ws):
        message = await ws.recv()
        data = json.loads(message)
        if data["action"] in ("subscribe", "establish"):
            await self.secondary(ws, data)
        else:
            await self.remote_primary(ws, message)

    async def main(self, ready=None):
        async with websockets.serve(self.remote_router, self.host, self.port), \
                websockets.unix_serve(self.local_router, self.server_unix_path):
            if ready is not None:
                ready.set()
            await asyncio.Future()

    def run(self, ready=None):
        asyncio.run(self.main(ready))

callback = FunctionList()

class ClientWS(WSHandshake):
    # TODO: maybe timeout after long silence and reconnect when a request hits

    def __init__(self, base_url, uri, cache=None, *, root_path=None):
        super().__init__(root_path)
        self.uri, self.cache = uri, cache
        self._url = base_url

    def url(self, path, query=None):
        # version via port numbers
        return self._url + "/login/ws" + path + (
            "" if query is None else "?" + urllib.urlencode(query))

    _public = None
    @property
    def public(self):
        if self._public is None:
            return self.load_public()
        return self._public

    @public.setter
    def public(self, value):
        self._public = value

    def load_public(self):
        if super().load_public():
            return self.client_verify_ack(requests.post(
                self.url("/syn"), self.client_syn()).content)
        return self.public

    def update(self):
        since = self.db().queryone("SELECT MAX(revoked_time) FROM ignore")[0]
        response = requests.post(
            self.url("/updates"), self.otp(),
            params=since and {"since": str(since)})
        data = json.loads(response.content)
        self.revoke(data["data"])
        self.purge(*data["timeouts"])

    def revoke(self, info):
        # TODO: cache sync
        self.db().executemany(
            "INSERT INTO ignore(ref, revoked_time, refresh, refresh_time) "
            "VALUES (?, ?, ?, ?)", info)

    def purge(self, lease_timeout, refresh_timeout):
        now = time.time()
        if lease_timeout is not None:
            self.db().execute(
                "DELETE FROM ignore WHERE refresh_time - ?>?",
                (now, lease_timeout))

    async def io_hook(self, message, reply):
        # print(message, reply)
        data = json.loads(message)
        if data["action"] in callback:
            callback[data["action"]](self, data, reply)

    def router(self, message):
        return self.revoke(message["data"])

    async def listen(self, ready=None):
        q = asyncio.Queue()
        async def handler(ws):
            e = asyncio.Event()
            async for message in ws:
                await q.put((ws, message, e))
                await e.wait()

        async with websockets.connect(self.uri) as notify, \
                websockets.connect(self.uri) as query, \
                websockets.unix_serve(handler, self.client_unix_path):
            await notify.send(json.dumps({
                "action": "subscribe",
                "data": base64.b64encode(self.otp()).decode()}))
            await query.send(json.dumps({
                "action": "establish",
                "data": base64.b64encode(self.otp()).decode()}))
            remote, local = list(map(
                asyncio.create_task, [notify.recv(), q.get()]))
            await asyncio.to_thread(self.update)
            if ready is not None:
                ready.set()

            while True:
                done, pending = await asyncio.wait(
                    [remote, local], return_when=asyncio.FIRST_COMPLETED)
                done = next(iter(done))
                if remote is done:
                    self.router(json.loads(self.client_recieve(done.result())))
                    remote = asyncio.create_task(notify.recv())
                else:
                    res = done.result()
                    ws, message, event = done.result()
                    await query.send(self.client_send(message))
                    reply = self.client_recieve(await query.recv())
                    await self.io_hook(message, reply)
                    await ws.send(reply)
                    event.set()
                    local = asyncio.create_task(q.get())

    def run(self, ready=None):
        asyncio.run(self.listen(ready))

class ClientBP(Handshake):
    def _closure(self, f):
        sig = inspect.signature(f)
        sig = sig.replace(parameters=list(sig.parameters.values())[1:])
        @functools.wraps(f)
        def wrapper(*a, **kw):
            return self._send(f.__name__, **sig.bind(*a, **kw).arguments)
        return staticmethod(wrapper)

    def __init__(self, *a, local=False, root_path=None, **kw):
        super().__init__(root_path)
        self.a, self.kw, self.unix_path = a, kw, \
            self.server_unix_path if local else self.client_unix_path
        for k, v in actionable.items():
            setattr(self, k, self._closure(v))

    async def _send(self, action, **kw):
        async with websockets.unix_connect(self.unix_path) as websocket:
            await websocket.send(json.dumps({"action": action, **kw}))
            return json.loads(await websocket.recv())

    def _fork(self):
        return super()._fork(
            ClientWS(*self.a, root_path=self.root_path(), **self.kw))

class SessionFromValueMixin:
    bp = None # unimplemented

    async def __call__(self, *a, group=None, **kw):
        if callable(group) and not isinstance(group, AccessGroup):
            user = await self.bp.from_value(
                self.get_cookie(*a, **kw), self.remote_addr(*a, **kw))
            group = group(user)
        res = await self.bp.from_value(
            self.get_cookie(*a, **kw), self.remote_addr(*a, **kw), group)
        if res["modified"]:
            self.set_cookie(res["cookie"], *a, **kw)
        return res["user"]

    def remote_addr(self, *a, **kw):
        raise NotImplementedError()

    def get_cookie(self, *a, **kw):
        raise NotImplementedError()

    def set_cookie(self, value, *a, **kw):
        raise NotImplementedError()

class LocalLoginInterface(LoginBuilder, SessionFromValueMixin):
    def __init__(self, root_path=None):
        super().__init__(prefix=None, root_path=root_path)
        self.bp = ClientBP(local=True, root_path=root_path)

class RemoteLoginBuilder(LoginBuilder):
    def __init__(self, base_url, uri, app=None, root_path=None, g_attr="user"):
        app_kw = {} if app is None else {"app": app}
        super().__init__(
            prefix=base_url, g_attr=g_attr, root_path=root_path, **app_kw)
        self.bp = ClientBP(base_url=self.endpoint, uri=uri, root_path=root_path)
        self.lease_timeout, self.refresh_timeout = self.app.config.get(
            "TIMEOUTS", default_timeouts)

    def membership(self, group, user):
        return self.bp.access_query(user, group)

    _db = None
    @property
    def db(self):
        if self._db is None:
            self._db = FKDatabase(
                self.app, self.bp.run_path("users.db"), relpath("schema.sql"))
        return self._db

    def auth(self, redirect=None, required=True, session=None):
        # only checks timing, blacklist
        session, now = session or self.session, time.time()
        bounced = lambda: self.bounce(redirect) if required else None
        if "user" not in session:
            return bounced()
        if self.refresh_timeout and \
                now - session["authtime"] > self.refresh_timeout:
            session.clear()
            return bounced()
        # TODO: caching
        if self.db.queryone(
                "SELECT EXISTS(SELECT 1 FROM ignore WHERE refresh=?)",
                (session["refresh"],))[0]:
            session.clear()
            return bounced()
        if self.lease_timeout and \
                now - session["refresh_time"] > self.lease_timeout:
            refresh_time = asyncio.run(self.bp.refresh(session["refresh"]))
            if refresh_time is None:
                session.clear()
                return bounced()
            session["refresh_time"] = refresh_time
        return {
                "id": session["user"],
                "name": session["name"],
                "picture": session["picture"],
            }

class RemoteLoginInterface(RemoteLoginBuilder, SessionFromValueMixin):
    def __init__(self, base_url, uri, app=None, root_path=None):
        super().__init__(base_url, uri, app, root_path)
        self.session_loader = SessionLoader(root_path)

    _db = None
    @property
    def db(self):
        if self._db is None:
            self._db = ThreadDB(
                self.bp.run_path("users.db"), relpath("schema.sql"))
        return self._db

    async def __call__(self, *a, group=None, **kw):
        if group is None:
            session = self.session_loader.open(self.get_cookie(*a, **kw))
            res = self.auth(required=False, session=session)
            if session.modified:
                session = self.session_loader.save(cookie)
                self.set_cookie(session, *a, **kw)
            return res
        return await super().__call__(*a, group=group, **kw)

def repl(client):
    while True:
        message = input("send action [argv 0] JSON [rest]: ")
        if message == "":
            continue
        message = message.split(" ", 1)
        message = message if len(message) == 2 else message + ["{}"]
        if message[0] not in actionable:
            print("invalid action call")
            continue
        print(asyncio.run(client._send(message[0], **json.loads(message[1]))))

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'launching', choices=["client", "server", "all", "repl"])
    parser.add_argument('--host-url', default="http://localhost:8000")
    parser.add_argument('--ws-port', default=8001)
    parser.add_argument('--ws-host', default="localhost")
    args = parser.parse_args()

    server_factory = ServerWS if args.launching == "server" else ServerBP
    client_factory = ClientBP if args.launching == "repl" else ClientWS
    server = server_factory(host=args.ws_host, port=args.ws_port)
    client = client_factory(
        base_url=args.host_url, uri=f"ws://{args.ws_host}:{args.ws_port}")

    if args.launching == "server":
        server.run()
    elif args.launching in {"all", "repl"}:
        server._fork()
    if args.launching in {"client", "all"}:
        client.run()
    repl(client._fork())




import os.path, functools, collections, flask

def relpath(*args):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), *args)

project_path = lambda *a: relpath("..", "..", *a)

from flask_caching.backends.memcache import MemcachedCache

# https://github.com/memcached/memcached/wiki/ConfiguringServer#unix-sockets
# remember TLS for all sensitive ISP traffic, see: MUSCULAR

# TODO: why does this seemingly work without the server running?
class ThreadedMemcached(MemcachedCache):
    def import_preferred_memcache_lib(self, servers):
        import libmc
        return libmc.ThreadedClient(servers, hash_fn=libmc.MC_HASH_FNV1_32)

def threaded_client(app, config, args, kwargs):
    return ThreadedMemcached.factory(app, config, args, kwargs)

import json

def dict_names(o):
    if hasattr(o, "_asdict"):
        o = o._asdict()
    if any(isinstance(o, i) for i in (list, tuple)):
        return tuple(dict_names(i) for i in o)
    if isinstance(o, dict):
        return {k: dict_names(v) for k, v in o.items()}
    return o

class RouteLobby:
    def __init__(self):
        self.routes = []

    def route(self, *a, **kw):
        def wrapper(f):
            self.routes.append((a, kw, f))
            return f
        return wrapper

    def register_lobby(self, bp, *fa, **fkw):
        for a, kw, f in self.routes:
            bp.route(*a, **kw)(
                functools.wraps(f)(functools.partial(f, *fa, **fkw)))

    def template_json(self, rule, template_path, prefix="/view", **routeargs):
        def decorator(f):
            def json_wrapper(*a, **kw):
                res = f(*a, **kw)
                if isinstance(res, flask.Response):
                    # TODO: not really implied, sort of a work around
                    if 300 <= res.status_code < 400:
                        flask.abort(401)
                    return res
                return json.dumps(dict_names(res))
            def template(*a, **kw):
                res = f(*a, **kw)
                if isinstance(res, flask.Response):
                    return res
                return flask.render_template(template_path, **res)

            json_wrapper.__name__ = f.__name__ + "_json"
            template.__name__ = f.__name__ + "_template"

            self.route(rule, **routeargs)(json_wrapper)
            self.route(prefix + rule, **routeargs)(template)
            return f
        return decorator

key_paths = (project_path("run"), project_path())
def secret_key(paths = key_paths, key_name="login_secret_session_key"):
    if isinstance(paths, str):
        paths = (paths,)
    for path in paths:
        file = os.path.join(path, key_name)
        if os.path.exists(file):
            with open(file, "rb") as f:
                return f.read()

    os.makedirs(paths[0], exist_ok=True)
    with open(os.path.join(paths[0], key_name), "wb") as f:
        secret = os.urandom(24)
        f.write(secret)
    return secret

import math

class CompressedUUID:
    compressed = "23456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
    base16 = "0123456789abcdef"
    # [i for i, j in enumerate(str(uuid.uuid4())) if j == '-']
    dashes = [8, 13, 18, 23]

class CompressedUUID(CompressedUUID):
    length = math.ceil(math.log(16 ** 32) / math.log(
        len(CompressedUUID.compressed)))

    @staticmethod
    def rebase(value, inbase, outbase):
        output = []
        for remainder in value:
            for digit in range(len(output)):
                remainder = inbase * output[digit] + remainder
                output[digit] = remainder % outbase
                remainder = remainder // outbase
            while remainder:
                output.append(remainder % outbase)
                remainder = remainder // outbase
        return output[::-1]

    @classmethod
    def translate(cls, value, inalphabet, outalphabet):
        rebased = cls.rebase(map(lambda x: inalphabet.index(x), value),
            len(inalphabet), len(outalphabet))
        return "".join(map(lambda x: outalphabet[x], rebased))

    @classmethod
    def fromUUID(cls, strUUID):
        b16str = strUUID.replace('-', '')
        small = cls.translate(b16str, cls.base16, cls.compressed)
        return small.rjust(cls.length, cls.compressed[0])

    @classmethod
    def toUUID(cls, short):
        b16str = cls.translate(short, cls.compressed, cls.base16)
        for i in cls.dashes:
            b16str = b16str[:i] + "-" + b16str[i:]
        return b16str

    @classmethod
    def possible(cls, unknown):
        return len(unknown) == cls.length and all(
            i in cls.compressed for i in unknown)

# ensures that json value matches template given
# templates with one list element can be an arbitrary length
# templates with multiple list elements must have the elements match
# dictionaries are turned into `namedtuple`s sorted alphabetically
# templates with sets are considered enums, and value must be in that set
# other than a nested set, sets can contain all the other kinds of templates
# otherwise, the type of the value must match the type given in template
def data_payload(value, template, parsed=True):
    def oxford_comma(terms):
        return " and ".join(terms) if len(terms) < 3 else \
            ", ".join(terms[:-1]) + ", and " + terms[-1]

    def ensure(payload, template, qualname=""):
        part_name = f"payload" + qualname
        requires = template if type(template) == type else type(template)
        if requires == set:
            assert len(template) > 0
            flag, has = type('flag', (), {})(), type(payload)
            it = ((flag, i) if type(i) == type else (i, flag) for i in template)
            values, types = map(lambda x: set(x).difference({flag}), zip(*it))
            if has in set(map(type, values)).intersection({dict, list}):
                for option in (i for i in values if type(i) == has):
                    try:
                        return ensure(payload, option, qualname)
                    except:
                        pass
            elif has in types or payload in values:
                return payload
            raise Exception(f"{part_name} has invalid value for enum")
        if not isinstance(payload, requires):
            raise Exception(f"{part_name} should be {requires}")
        if requires == dict:
            if template.keys() != payload.keys():
                given, needed = set(payload.keys()), set(template.keys())
                missing = oxford_comma(needed.difference(given))
                extra = oxford_comma(given.difference(needed))
                # one value can not be both missing and extra
                xor = {missing: "is missing ", extra: "should not contain "}
                message = part_name + " " + " and ".join(
                    v + k for k, v in xor.items() if k)
                raise Exception(message)
            ordered_names = tuple(sorted(template.keys()))
            obj = collections.namedtuple(
                part_name.replace(".", "__"), ordered_names)
            return obj(**{
                k: ensure(payload[k], template[k], qualname + f".{k}")
                for k in ordered_names})
        elif requires == list:
            idx = (lambda i: 0) if len(template) == 1 else (lambda i: i)
            return [
                ensure(v, template[idx(i)], qualname + f"_{i}")
                for i, v in enumerate(payload)]
        else:
            return payload

    if parsed:
        payload = value
    else:
        try:
            payload = json.loads(value)
        except json.decoder.JSONDecodeError:
            flask.abort(415, description="invalid JSON")
    try:
        return ensure(payload, template)
    except Exception as e:
        # raise e
        flask.abort(400, description=e.args[0])

class OpShell:
    _op_on, _op, _op_args = None, None, ()

    symbols = (
        ("_and", lambda l, r: f"({l} AND {r})"),
        ("_or", lambda l, r: f"({l} OR {r})"),
        ("_xor", lambda l, r: f"({l} XOR {r})"),
        ("_invert", lambda x: f"NOT {x}"),
    )

    @property
    def _shell(self):
        return __class__

    def __init__(self, this=None, op=None, *a):
        self._op_on, self._op, self._op_args = this, op, a
        self._op_symbols = {
            getattr(self.__class__, k): v for k, v in self.symbols}

    @property
    def _on(self):
        return self if self._op_on is None else self._op_on

    def __contains__(self, user):
        return bool(self.generic("__contains__")(user))

    def __getattr__(self, name):
        return self.generic(name)

    def generic(self, f, *a, **kw):
        if self._op is None:
            assert getattr(self._on.__class__, f, None) != getattr(
                self._shell, f, None)
        def wrapper(*a, **kw):
            if self._op is None:
                return getattr(self._on, f)(*a, **kw)
            return self._op(f, *self._op_args, *a, **kw)
        return functools.wraps(f)(wrapper)

    def __pos__(self):
        return abs(self) if self._op is None else self

    def __neg__(self):
        return self if self._op is None else abs(self)

    def __abs__(self):
        return self._shell(self, self._op, *self._op_args)

    def _and(self, f, other, *a, **kw):
        return getattr(self, f)(*a, **kw) and getattr(other, f)(*a, **kw)

    def __and__(self, other):
        return self._shell(self, self._and, other)

    def __iand__(self, other):
        this = +self
        this._op, this._op_args = (-this)._and, (other,)
        return this

    def _or(self, f, other, *a, **kw):
        return getattr(self, f)(*a, **kw) or getattr(other, f)(*a, **kw)

    def __or__(self, other):
        return self._shell(self, self._or, other)

    def __ior__(self, other):
        this = +self
        this._op, this._op_args = (-this)._or, (other,)
        return this

    def _xor(self, f, other, *a, **kw):
        return bool(
            getattr(self, f)(*a, **kw) ^ getattr(other, f)(*a, **kw))

    def __xor__(self, other):
        return self._shell(self, self._xor, other)

    def __ixor__(self, other):
        this = +self
        this._op, this._op_args = (-this)._xor, (other,)
        return this

    def _invert(self, *a, **kw):
        return not getattr(self, f)(*a, **kw)

    def __invert__(self):
        return self._shell(self, self._invert)

    def __repr__(self):
        if self._op is None:
            return super().__repr__() if self._on is self else repr(self._on)
        return self._op_symbols[self._op.__func__](
            repr(self._on), *map(repr, self._op_args))

    def rpn(self):
        if self._op is None:
            return super().__str__() if self._on is self else str(self._on)
        return [
            f"__{self._op.__func__.__name__.replace('_', '')}__",
            *map(lambda x: x.rpn(), [self._on, *self._op_args])]

    def __str__(self):
        return json.dumps(self.rpn())

class FlatInfo(tuple):
    pass

class InfoShell(OpShell):
    @property
    def _shell(self):
        return __class__

    def __init__(self, this=None, op=None, *a):
        # TODO: could be distributive
        assert op.__func__.__name__ == "_add" or (
            this._op is None or this._op.__func__.__name__ != "_add") and (
            not a or a[0]._op is None or
            a[0]._op.__func__.__name__ != "_add")
        self.symbols += ("_add", lambda l, r: f"({l} ADD {r})"),
        super().__init__(this, op, *a)

    def _add(self, f, other, *a, **kw):
        res = (getattr(self, f)(*a, **kw), getattr(other, f)(*a, **kw))
        return FlatInfo(
            j for i in res for j in (i if isinstance(i, FlatInfo) else (i,)))

    def __add__(self, other):
        return self._shell(self, self._add, other)

    def __iadd__(self, other):
        this = +self
        this._op, this._op_args = (-this)._add, (other,)
        return this

class OpShellTest(OpShell):
    def __init__(self, lo, hi):
        self.lo, self.hi = lo, hi

    def __contains__(self, value):
        return self.lo <= value <= self.hi

    def test(self, x):
        return x not in self

    def __repr__(self):
        return f"{self.lo} to {self.hi}"

# a = OpShellTest(2, 3) & OpShellTest(3, 4)
# b = OpShellTest(2, 3) | OpShellTest(3, 4)
# c = OpShellTest(7, 8)
# d = c
# d |= a

def endpoint_next(endpoint, redirect=None):
    return endpoint + "?" + urllib.parse.urlencode(
        {"next": redirect or flask.request.url})




import json, time, secrets, uuid, urllib.parse
import flask
from flask_dance.consumer import oauth_before_login, oauth_authorized
from flask_dance.consumer.storage import BaseStorage
from flask_dance.utils import FakeCache

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from platforms import methods, userlookup

end_locals()

def safe_redirect(url, app=None):
    app = flask.current_app if app is None else app
    checking = urllib.parse.urlparse(url)
    # fallback could be app.config['SERVER_NAME'] but if multiple subdomains
    # point endpoints to the same flask app (same API token, different logins)
    # then the url redirect should be within the same host
    trusted = app.config['SESSION_COOKIE_DOMAIN'] or flask.request.host
    # cookies shared across ports
    trimmed = checking.netloc.rsplit(":")[0]
    trusted = trusted.rsplit(":")[0]
    valid = not trimmed or trimmed.endswith(trusted)
    if "localhost" in trusted:
        return valid
    return checking.scheme in ("", "https", "wss") and valid

def get_next():
    try:
        return json.loads(flask.session.get("next", "{}"))
    except json.JSONDecodeError:
        return {}

def store_next(stored):
    flask.session["next"] = json.dumps(stored)

# TODO: customizable fallback URL ("/" right now)
@oauth_before_login.connect
def before_login(blueprint, url):
    state = urllib.parse.parse_qs(urllib.parse.urlparse(url)[4])["state"][0]
    stored = get_next()
    current_redirect = flask.request.args.get("next", "/")
    if not safe_redirect(current_redirect):
        flask.abort(400)
    stored[state] = current_redirect
    store_next(stored)

@oauth_authorized.connect
def logged_in(blueprint, token):
    state = flask.request.args["state"]
    stored = get_next()
    next_url = stored.pop(state, "/") # TODO: fallback URL here too
    store_next(stored)
    blueprint.token = token
    return flask.redirect(next_url)

def authorized(session_=None):
    session_ = session_ or flask.session
    method = session_.get("method", None)
    if method in methods:
        return methods[method][0].authorized

default_timeouts = (3600 * 24, None)

# this needs to go through login server for remote calls
def refresh_access(db, refresh, refresh_time, lease_timeout, cached):
    now = int(time.time())
    if lease_timeout and now - refresh_time > lease_timeout:
        if cached:
            update = db.queryone(
                "SELECT refresh_time from active WHERE refresh=?",
                (refresh,))
            if update is not None and now - update[0] < lease_timeout:
                return True, False, update[0]
        db.execute(
            "UPDATE active SET refresh_time=? WHERE refresh=?",
            (now, refresh))
        return True, True, now
    return False, False, refresh_time

class DBStore(BaseStorage):
    def __init__(
            self, db, method, cache=None, session_=None,
            lease_timeout=default_timeouts[0],
            refresh_timeout=default_timeouts[1]):
        super().__init__()
        self.db, self.method = db, method
        self.lease_timeout = lease_timeout
        self.refresh_timeout = refresh_timeout
        self.cache = cache or FakeCache()
        self.session = session_ or (lambda: flask.session)

    def set(self, blueprint, token):
        session_ = self.session()
        session_["method"] = self.method
        info = userlookup(self.method)
        encoded = json.dumps(token)
        uniq = str(uuid.uuid4())

        # upsert returns None
        uid = self.db.execute(
            "INSERT INTO auths"
            "(method, platform_id, display_name, picture, token, uuid) "
            "VALUES (?, ?, ?, ?, ?, ?) ON CONFLICT(method, platform_id) "
            "DO UPDATE SET "
            "token=excluded.token, "
            "display_name=excluded.display_name, "
            "picture=excluded.picture",
            (self.method, info["id"], info["name"], info["picture"],
             encoded, uniq))
        if uid is None:
            uniq = uid if uid is not None else self.db.queryone(
                "SELECT uuid FROM auths WHERE method = ? AND platform_id = ?",
                (self.method, info["id"]))[0]

        authtime = int(time.time())
        refresh = secrets.token_urlsafe(32)
        ip = flask.request.remote_addr
        session_["user"], session_["authtime"] = uniq, authtime
        session_["refresh"], session_["refresh_time"] = refresh, authtime
        session_["name"], session_["picture"] = info["name"], info["picture"]
        self.db.execute(
            "INSERT INTO active"
            "(uuid, refresh, ip, authtime, refresh_time) "
            "VALUES (?, ?, ?, ?, ?)",
            (uniq, refresh, ip, authtime, authtime))
        self.cache.set(refresh, (encoded, ip, authtime, authtime))

    def get(self, blueprint):
        db = self.db.begin()
        session_, info = self.session(), None
        if "refresh" not in session_:
            return None

        refresh = session_["refresh"]
        cached = self.cache.get(refresh)
        if cached is None:
            info = db.queryone(
                "SELECT auths.token, active.ip, "
                "active.authtime, active.refresh_time "
                "FROM active LEFT JOIN auths ON auths.uuid=active.uuid "
                "WHERE active.refresh=?", (refresh,))
            if info is None:
                session_.clear()
                return None

            token, ip, authtime, refresh_time = info
            info = list(info)
        else:
            token, ip, authtime, refresh_time = cached
            cached = list(cached)

        if self.refresh_timeout and int(
                time.time()) - authtime > self.refresh_timeout:
            self.deauthorize(refresh)
            session_.clear()
            return None

        updated, write, refresh_time = refresh_access(
            db, refresh, refresh_time, self.lease_timeout,
            cached is not None)

        if updated:
            info = info or cached
            info[3] = refresh_time
            session_["refresh_time"] = refresh_time

        current_ip = flask.request.remote_addr
        if ip != current_ip:
            db.execute("UPDATE active SET ip=? WHERE refresh=?",
                (current_ip, refresh))
            info = info or cached
            info[1] = current_ip
            write = True

        if info is not None:
            self.cache.set(refresh, tuple(info))

        if write:
            db.commit()
        db.close()

        return json.loads(token)

    def delete(self, blueprint):
        session_ = self.session()
        if "user" not in session_:
            return None

        for (refresh,) in self.db.queryall(
                "SELECT refresh FROM active WHERE uuid=?", (session_["user"],)):
            self.deauthorize(refresh)

        self.db.execute("DELETE FROM auths WHERE uuid=?", (session_["user"],))

    def deauthorize(self, refresh, user=None, callback=lambda *a: None):
        session_ = self.session()
        db = self.db.begin()
        user_query, user_args = ("", ()) if user is None else \
            (" AND user=?", (user,))
        authtime = db.queryone(
            "SELECT authtime FROM active WHERE refresh=?" + user_query,
            (refresh,) + user_args)
        if authtime == None:
            return None
        now = float(time.time())
        rowid = db.execute(
            "INSERT INTO revoked(revoked_time, refresh, authtime, "
            "refresh_time) VALUES (?, ?, ?, ?)", (
                now, session_["refresh"], authtime[0],
                session_["refresh_time"]))
        callback(
            rowid, now, session_["refresh"], session_["refresh_time"])
        db.execute(
            "DELETE FROM active WHERE refresh=?", (refresh,))
        self.cache.delete(refresh)
        db.commit().close()




from flask_dance.contrib.google import google, make_google_blueprint
from flask_dance.contrib.facebook import facebook, make_facebook_blueprint
from flask_dance.contrib.github import github, make_github_blueprint

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from test import test, make_test_blueprint

end_locals()

methods = {
    "google": (google, make_google_blueprint, None),
    "facebook": (facebook, make_facebook_blueprint, None),
    "github": (github, make_github_blueprint, None),
    "test": (test, make_test_blueprint, None),
}

# returns {"id": "...", "name": "...", "picture": "..."}
def userlookup(method):
    if method == "google":
        return remap(google.get("/oauth2/v3/userinfo").json(), {
            "id": "sub", "name": "name", "picture": "picture"})
    elif method == "facebook":
        conf = ".width(200).height(200)"
        return remap(facebook.get(f"/me?fields=id,name,picture{conf}").json(), {
            "id": "id", "name": "name", "picture": ["picture", "data", "url"]})
    elif method == "github":
        return remap(github.get("/user").json(), {
            "id": "node_id", "name": "name", "picture": "avatar_url"})
    elif method == "test":
        return test.get()

# remapping to an array accesses nested dicts
def remap(old, mapping):
    res = {}
    for k, v in mapping.items():
        r = old
        for i in [v] if type(v) == str else v:
            r = r if r is None else r.get(i, None)
        res[k] = r
    return res




import datetime, flask, urllib.parse, functools, time

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from interface import OAuthBlueprint
from tokens import authorized
from utils import RouteLobby
from pubsub import ServerBP

end_locals()

login_lobby = RouteLobby()

# all endpoints can only be called with flask.current_app as the auth app
class LoginBlueprint(OAuthBlueprint):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        login_lobby.register_lobby(self, self)
        self._oauth_ws = ServerBP(
            self._oauth_db, root_path=self._oauth_root_path)
        self.register_blueprint(self._oauth_ws.bp)

    def _oauth_deauthorize(self, token, method):
        super()._oauth_deauthorize(
            token, method, callback=self._oauth_ws.deauthorize)

    @login_lobby.route("/logout")
    def _oauth_logout(self):
        if "method" in flask.session:
            self._oauth_deauthorize(
                flask.session["refresh"], flask.session["method"])
            flask.session.clear()
        return flask.redirect(flask.request.args.get(
            "next", self.login_endpoint()))

    @login_lobby.route("/deauthorize/<refresh>", methods=["POST"])
    def _oauth_kick(self, refresh):
        if not authorized():
            flask.abort(401)
        self._oauth_deauthorize(
            refresh, flask.session["method"], flask.session["user"])

    @login_lobby.template_json("/sessions", "sessions.html")
    def _oauth_sessions(self):
        if not authorized():
            return flask.redirect(
                self.login_endpoint() + "?" + urllib.parse.urlencode(
                    {"next": flask.request.url}))
        active = [
            dict(zip(["token", "ip", "authtime", "refresh_time"], sess))
            for sess in self._oauth_db().queryall(
                "SELECT refresh, ip, authtime, refresh_time FROM active "
                "WHERE uuid = ?", (flask.session["user"],))]
        for sess in active:
            sess["authtime"] = datetime.datetime.fromtimestamp(
                sess["authtime"]).strftime("%m/%d/%Y %H:%M:%S UTC")
            sess["current"] = sess["token"] == flask.session["refresh"]
        return {"active": active}




import flask
from werkzeug.local import LocalProxy
from functools import wraps

class MonoBlueprint(flask.Blueprint):
    apps = []

    def register(self, app, options):
        super().register(app, options)
        self.apps.append(app)

class TestBP(MonoBlueprint):
    def __init__(self):
        super().__init__("test", __name__)
        self.route("/test")(self.debug_only(login))
        self.route("/test/as")(self.debug_only(self.test_auth_as))
        self.route("/test/failure")(self.debug_only(failure))

    def debug_only(self, f):
        @wraps(f)
        def wrapped(*a, **kw):
            if not all(app.debug for app in self.apps):
                flask.abort(403)
            return f(*a, **kw)
        return wrapped

    def test_auth_as(self):
        who = test_whois()
        if len(who) == 0:
            flask.abort(403)
        test.store.set(self, who)
        # TODO: customizable fallback
        return flask.redirect(flask.request.args.get("next", "/"))

def login():
    url = {"next": flask.request.args["next"]} if "next" in flask.request.args \
        else {}
    url = flask.url_for("test.test_auth_as", **url)
    return flask.render_template(
        "test.html", ip=flask.request.remote_addr, next=url)

def failure():
    raise RuntimeError()

test_whois = lambda: flask.request.args.get("who", "")
test_pic = "https://github.githubassets.com/assets/GitHub-Mark-ea2971cee799.png"

class TestMockSession():
    def __init__(self, store):
        self.store = store

    @property
    def authorized(self):
        return self.store.get(self)

    def get(self):
        id = test_whois()
        return {"id": id, "name": id, "picture": test_pic}

def make_test_blueprint(storage, **kw):
    test_session = TestMockSession(storage)
    test_bp = TestBP()

    @test_bp.before_app_request
    def set_applocal_session():
        flask.g.flask_dance_test = test_session
    return test_bp

test = LocalProxy(lambda: flask.g.flask_dance_test)




import os.path, functools, collections, threading

# http://flask.pocoo.org/docs/0.11/patterns/sqlite3/
import sqlite3, contextlib
import flask

class Passthrough:
    rewrite = ()

    def __init__(self, parent):
        self.parent = parent

    def __getattribute__(self, name):
        if name in ("parent", "rewrite") or name in self.rewrite:
            return object.__getattribute__(self, name)
        return getattr(self.parent, name)

class TransactionCursor(Passthrough):
    rewrite = ("close",)

    def close(self):
        pass

class TransactionConnection(Passthrough):
    rewrite = ("commit", "cursor", "_transaction_cursor")

    def __init__(self, parent):
        super().__init__(parent)
        self._transaction_cursor = None

    def cursor(self):
        if self._transaction_cursor is None:
            self._transaction_cursor = TransactionCursor(self.parent.cursor())
        return self._transaction_cursor

    def commit(self):
        pass

class TransactionContext(Passthrough):
    rewrite = ("get", "commit", "close", "_transaction_con", "begin", "ctx")

    def __init__(self, parent):
        super().__init__(parent)
        self._transaction_con = None

    def __getattribute__(self, name):
        res = super().__getattribute__(name)
        if name not in ("parent", "rewrite") and name not in self.rewrite:
            member = getattr(self.parent, name, None)
            if callable(member):
                return member
        return res

    def get(self):
        if self._transaction_con is None:
            self._transaction_con = TransactionConnection(self.parent.get())
        return self._transaction_con

    def commit(self):
        self.parent.commit()
        return self

    def close(self):
        if self._transaction_con is not None:
            if self._transaction_con._transaction_cursor is not None:
                self._transaction_con._transaction_cursor.close()
        if isinstance(self.parent, AppContext):
            self.parent.close()

    @property
    def ctx(self):
        raise Exception("context should be added before the transaction")

class AppContext(Passthrough):
    rewrite = ("wrapper", "ctx", "begin", "_transacting")

    def __init__(self, parent):
        super().__init__(parent)
        self._transacting = False

    def __getattribute__(self, name):
        res = super().__getattribute__(name)
        return self.wrapper(res) if \
            name not in ("parent", "rewrite") and \
            name not in self.rewrite and callable(res) \
            else res

    def wrapper(self, f):
        @functools.wraps(f)
        def wrapped(*a, **kw):
            return self.ctx(lambda: f(*a, **kw))
        return wrapped

    def ctx(self, f):
        if self._transacting:
            return f()
        with self.parent.app.app_context():
            return f()

    def begin(self):
        self._transacting = True
        self.parent.app.app_context().__enter__()
        return TransactionContext(self)

    def close(self):
        self._transacting = False
        self.parent.app.app_context().__exit__()

def split_out(flat, seps):
    flat = (flat,) if type(flat) == str else flat
    for sep in seps:
        flat = tuple(filter(None, sum((sum((
            [j, sep] for j in i.split(sep)), [])[:-1] for i in flat), [])))
    return flat

def seek_layer(iterator, pause, done):
    parens = 0
    for i in iterator:
        if i == "(":
            parens += 1
        elif i == ")":
            parens -= 1
        elif parens != 0:
            continue
        elif i in pause:
            return True
        elif i in done:
            return False
    assert False, "invalid expression"

def sql_names(query, values, rowid=None):
    # https://www.sqlite.org/lang_select.html
    endings = {
        "from", "where", "group", "having", "window", "order", "limit",
        "union", "intersect", "except"}
    # https://www.sqlite.org/syntax/expr.html
    exprs = {
        "null", "true", "false", "current_time", "current_date",
        "current_timestamp", "is", "not", "and", "or", "in", "match",
        "like", "regexp", "glob", "collate", "isnull", "notnull",
        "between", "case", "cast", "raise"}

    words = split_out(query.lower().split(), ",()")
    cols, parens, staggered = [], 0, zip(
        *((None,) * (3 - i) + words + (None,) * i for i in range(4)))
    latest, continues = (i for _, _, _, i in staggered), True
    seek_layer(latest, "", ("select",))
    if next(latest) in ("distinct", "all"):
        next(latest)
    while continues:
        continues = seek_layer(latest, ",", endings)
        prev, word, _, _ = next(staggered)
        assert prev in ("as", ",", "distinct", "latest", "select")
        assert ord('a') <= ord(word[0]) <= ord('z')
        assert all(i not in word for i in "()*'\"- ")
        assert word not in exprs
        cols.append(word.rsplit(".", 1)[-1]) # TODO

    assert len(cols) == len(values)
    assert len(cols) == len(set(cols))
    obj = collections.namedtuple(
        "row" + ("" if rowid is None else str(rowid)), cols)
    return obj(**dict(zip(cols, values)))

class HeadlessDB:
    def __init__(self, database, schema, init=[], debug=False):
        self.database = os.path.abspath(database)
        self.schema, self.init, self.debug = schema, init, debug

        self._g, self.app = None, type("resource_app", (), {
            "open_resource": open, "app_context": contextlib.contextmanager(
                lambda: iter([None]))})()
        self.many = [self.queryone, self.queryall]

    def ensure(self):
        if not os.path.exists(self.database):
            with self.app.app_context():
                db = self.get()
                cur = db.cursor()
                with self.app.open_resource(self.schema, mode='r') as f:
                    cur.executescript(f.read())
                cur.close()
                db.commit()
                self.db_init_hook()

    @property
    def g(self):
        if self._g is None:
            self._g = type("global_store", (), {})()
        return self._g

    # returns a database connection
    def get(self):
        db = getattr(self.g, "_auth_database", None)
        if db is None:
            db = self.g._auth_database = {}
        if self.database not in db:
            con = db[self.database] = sqlite3.connect(self.database)
            for i in self.init:
                self.execute(i)
            # if False:
            if self.debug:
                con.set_trace_callback(print)
        return db[self.database]

    # TODO: add paging or switch to iterator before using at scale
    def queryall(self, query, args=(), names=False):
        cur = self.get().cursor()
        cur.execute(query, args)
        rv = cur.fetchall()
        cur.close()
        if names:
            return [sql_names(query, r) for r in rv]
        return rv

    def queryone(self, query, args=(), names=False):
        cur = self.get().cursor()
        cur.execute(query, args)
        rv = cur.fetchone()
        row = cur.lastrowid
        cur.close()
        if names and rv:
            return sql_names(query, rv, row)
        return rv

    def execute(self, query, args=()):
        con = self.get()
        cur = con.cursor()
        cur.execute(query, args)
        con.commit()
        res = cur.lastrowid
        cur.close()
        return res or None

    def executemany(self, query, args=()):
        con = self.get()
        cur = con.cursor()
        cur.executemany(query, args)
        con.commit()
        cur.close()

    def begin(self):
        return TransactionContext(self)

    def commit(self):
        return self.get().commit()

    def close(self):
        db = getattr(self.g, '_auth_database', None)
        if db is not None:
            for con in db.values():
                con.close()
            self.g._auth_database = {}

    def db_init_hook(self):
        pass

class Database(HeadlessDB):
    # creates database if it doesn't exist; set up by schema
    def __init__(self, app, database, schema, init=[], debug=False):
        super().__init__(database, schema, init, debug)
        self.app, self._g = app, flask.g
        if app != flask.current_app:
            app.teardown_appcontext(lambda e: self.close())
        self.ensure()

    @property
    def ctx(self):
        return AppContext(self)

class DefaultsDB:
    default_sql = []

    def get(self):
        if not hasattr(self.g, "_auth_database"):
            lower_init = set(sql.lower() for sql in self.init)
            prepend = []
            for sql in self.default_sql:
                if sql.lower() not in lower_init:
                    prepend.append(sql)
            self.init = prepend + self.init
        return super().get()

class FKDefault(DefaultsDB):
    default_sql = ["PRAGMA foreign_keys = ON"]

class FKDatabase(FKDefault, Database):
    pass

class FKHeadless(FKDefault, HeadlessDB):
    pass

class ThreadDB(FKHeadless):
    @property
    def g(self):
        g = super().g
        if not hasattr(g, "threads"):
            g.threads = {}
        t = g.threads
        tid = threading.current_thread().ident
        if tid not in t:
            t[tid] = type("thread_store", (), {})()
        return t[tid]




from .server import login_required, login_optional, auth_bp, AccessNamespace
from .builder import LoginBuilder, LoginBuilder as LocalLoginBuilder
from .endpoints import LoginBlueprint
from .pubsub import (
    RemoteLoginBuilder,
    RemoteLoginInterface,
    LocalLoginInterface
)




import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from builder import app, LoginCaller
from endpoints import LoginBlueprint

end_locals()

login_config = LoginCaller()
login_required = login_config.login_required
login_optional = login_config.login_optional

auth_bp = LoginBlueprint()
AccessNamespace = auth_bp.group
app.register_blueprint(auth_bp)

if __name__ == "__main__":
    app.run(port=8000)




import functools, urllib.parse
from werkzeug.middleware.proxy_fix import ProxyFix
import flask

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from utils import secret_key, endpoint_next
from tokens import authorized
from access import AccessGroup
from interface import OAuthBlueprint

end_locals()

# create the minimal app context so that other apps can push it to the stack and
# check the login status of the request
app = flask.Flask(__name__)
app.config["SESSION_COOKIE_NAME"] = "login"
app.wsgi_app = ProxyFix(app.wsgi_app)
# app.config["TIMEOUTS"] = (3600 * 24, 3600 * 24 * 90)

class PermissionsWrapper(tuple):
    pass

class LoginBuilder:
    def __init__(self, app=app, prefix=None, g_attr="user", root_path=None):
        self.app = app
        self.prefix = prefix
        self.g_attr = g_attr
        self.root_path = root_path
        root_path = () if root_path is None else (root_path,)
        app.secret_key = secret_key(*root_path)

    @property
    def endpoint(self):
        prefix = flask.request.root_url if self.prefix is None and \
            flask.current_app == self.app else self.prefix or ""
        return prefix + OAuthBlueprint.login_endpoint(self.app)

    @property
    def g(self):
        return flask.g.get(self.g_attr)

    @g.setter
    def g(self, value):
        flask.g.__setattr__(self.g_attr, value)

    @property
    def session(self):
        return OAuthBlueprint.session(self.app)

    def bounce(self, redirect=None, group=None):
        if group is not None:
            # TODO: request access page? from whom?
            flask.abort(403)
        if flask.request.method == "GET":
            return flask.redirect(endpoint_next(self.endpoint, redirect))
        else:
            flask.abort(401)

    def auth(self, redirect=None, required=True):
        session_ = self.session
        with self.app.app_context():
            # the flask dance blueprints modify the current context
            # with before_app_request for all requests to allow lookup
            for ctx_setup in flask.current_app.before_request_funcs[None]:
                ctx_setup()
            if not authorized(session_):
                if not required:
                    return
                return self.bounce(redirect)
            return {
                    "id": session_["user"],
                    "name": session_["name"],
                    "picture": session_["picture"],
                }

    def membership(self, group, user):
        return group.vet(user)

    def vet(self, user, group, redirect=None, required=True, kw=None):
        if callable(group) and not isinstance(group, AccessGroup):
            group = group(**kw)
        permissions = self.membership(group, user["id"])
        # TODO: cache
        if permissions is None and required:
            return self.bounce(redirect, group)
        if self.g is not None:
            permissions = PermissionsWrapper(sum(map(
                lambda x: x if isinstance(x, PermissionsWrapper) else (x,),
                (self.g["via"], permissions)), PermissionsWrapper()))
        return {**user, "via": permissions}

    def decorate(self, kw=None, required=True, group=None, redirect=None):
        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                user = self.auth(redirect, required)
                if user is None:
                    return f(*args, **kwargs)
                elif not isinstance(user, dict):
                    return user
                if group is not None:
                    group_args = {**({kw: user} if kw else {}), **kwargs}
                    user = self.vet(user, group, redirect, required, group_args)
                    if not isinstance(user, dict):
                        return user
                if kw is None or kw in kwargs and self.g is not None and \
                        kwargs[kw] != self.g:
                    self.g = user
                    return f(*args, **kwargs)
                self.g = user
                return f(*args, **{**kwargs, **({kw: user} if kw else {})})
            return wrapper
        return decorator

    def before_request(self, bp, required=True, group=None, redirect=None):
        def user_auth():
            res = self.auth(redirect, required)
            if not isinstance(res, dict):
                return res
            elif group is not None:
                res = self.vet(
                    res, group, redirect, required, kw=flask.request.view_args)
                if not isinstance(res, dict):
                    return res
            self.g = res
        bp.before_request(user_auth)

    def login_merged(
            self, ambiguous=None, kw=None, group=None, redirect=None,
            required=True):
        if isinstance(ambiguous, AccessGroup):
            if group is not None:
                raise TypeError(
                    f"{__class__.__name__}.login_*()"
                    " got multiple values for argument 'kw'")
            group, ambiguous = ambiguous, None
        if isinstance(ambiguous, str) or ambiguous is None or kw is not None:
            if ambiguous is not None and kw is not None:
                raise TypeError(
                    f"{__class__.__name__}.login_*()"
                    " got multiple values for argument 'kw'")
            kw = ambiguous if kw is None else kw
            return self.decorate(kw, required, group, redirect)
        elif isinstance(ambiguous, flask.Blueprint) or \
                isinstance(ambiguous, flask.Flask):
            self.before_request(ambiguous, required, group, redirect)
            return ambiguous
        elif callable(ambiguous):
            return self.decorate(None, required, group, redirect)(ambiguous)

    def login_required(
            self, ambiguous=None, kw=None, group=None, redirect=None):
        return self.login_merged(ambiguous, kw, group, redirect, True)

    def login_optional(
            self, ambiguous=None, kw=None, group=None, redirect=None):
        return self.login_merged(ambiguous, kw, group, redirect, False)

    @property
    def decorators(self):
        return self.login_required, self.login_optional

class BoundCall:
    caller, f = None, None

    @staticmethod
    def calls(caller, f):
        self = BoundCall()
        self.caller = caller
        self.f = f
        return self

    @property
    def login_required(self):
        return __class__.calls(self, super().login_required)

    @property
    def login_optional(self):
        return __class__.calls(self, super().login_optional)

    def __getattribute__(self, attr):
        if attr in ("caller", "f"):
            try:
                return object.__getattribute__(self, attr)
            except AttributeError:
                return None
        if object.__getattribute__(self, "caller") is not None:
            return getattr(self.caller, attr)
        return object.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        if attr in ("caller", "f") or self.caller is None:
            return object.__setattr__(self, attr, value)
        return setattr(self.caller, attr, value)

    def __call__(self, *args, **kwargs):
        assert self.f is not None, "invalid call to LoginBuilder base"
        return self.f(*args, **kwargs)

class LoginCaller(BoundCall, LoginBuilder):
    pass




import json, os.path, functools
import flask, flask_caching

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from store import FKDatabase, FKHeadless
from utils import RouteLobby, relpath, project_path
from tokens import DBStore, methods, authorized, safe_redirect
from access import AccessRoot

end_locals()

oauth_lobby = RouteLobby()

class OAuthBlueprint(flask.Blueprint):
    _oauth_name = "modular_login"
    _credentials_paths = ((), ("run",))
    _oauth_run_path = ("run",)
    # primarily for the sake of access group HTTP api endpoints
    _version = None

    def __init__(
            self, name=None, import_name=None, static_folder=None,
            static_url_path=None, template_folder=None, url_prefix=None,
            subdomain=None, url_defaults=None, root_path=None, cli_group=None):
        name = self._oauth_name if name is None else name
        import_name = __name__ if import_name is None else import_name
        root_path = project_path() if root_path is None else root_path
        url_prefix = "/login" if url_prefix is None else url_prefix
        if self._version is not None:
            url_prefix += "/" + self._version
        super().__init__(
            name, import_name, static_folder, static_url_path, template_folder,
            url_prefix, subdomain, url_defaults, root_path, cli_group)

        self._oauth_root_path = root_path
        self._oauth_run_root = os.path.join(root_path, *self._oauth_run_path)
        self._oauth_apps = []
        self._get_oauth_keys()
        oauth_lobby.register_lobby(self, self)
        self.record(lambda setup_state: self._oauth_register(setup_state.app))

        self.group = AccessRoot(self._oauth_db, OAuthBlueprint.login_endpoint)
        self.register_blueprint(self.group.bp)

    def _get_oauth_keys(self):
        for rel in self._credentials_paths:
            path = os.path.join(self._oauth_run_root, *rel, "credentials.json")
            if os.path.exists(path):
                with open(path) as f:
                    self._oauth_keys = json.load(f)
                return
        self._oauth_keys = {
            name: {"id": "", "secret": ""} for name in methods.keys()}

    @oauth_lobby.route("/")
    def login_base_route(self):
        url = flask.request.args.get("next")
        if url is not None and not safe_redirect(url):
            flask.abort(400)
        if not authorized(): # only accessed in auth app's context
            url = {} if url is None else {"next": url}
            debug = all(i.debug for i in self._oauth_apps)
            # method.login is a flask-dance interface
            return flask.render_template("login.html", debug=debug, **{
                name: flask.url_for(method.name + ".login", **url)
                for name, method in self._oauth_blueprints.items()})
        # TODO: allow custom fallback endpoint
        return flask.redirect("/" if url is None else url)

    def _oauth_deauthorize(self, token, method, callback=None):
        self._oauth_stores[method].deauthorize(token, callback=callback)

    def _oauth_db(self, app=None):
        app = app or flask.current_app
        db_path = os.path.join(self._oauth_run_root, "users.db")
        if app:
            return FKDatabase(
                app, db_path, relpath("schema.sql"), debug=app.debug)
        else:
            return FKHeadless(db_path, relpath("schema.sql"))

    def _oauth_register(self, app):
        self._oauth_apps.append(app)
        db = self._oauth_db(app)
        cache = flask_caching.Cache(app, config={
            'CACHE_MEMCACHED_SERVERS': [project_path("run", "memcached.sock")],
            'CACHE_TYPE': 'utils.threaded_client'})
        stores, blueprints = {}, {}
        # keep version out of redirect since that's provider facing
        for name, (_, factory, scope) in methods.items():
            stores[name] = DBStore(
                db, name, cache, lambda: self.session(app),
                *app.config.get("TIMEOUTS", ()))
            blueprints[name] = factory(
                client_id=self._oauth_keys[name]["id"],
                client_secret=self._oauth_keys[name]["secret"],
                redirect_url=f"/login/{name}/continue",
                storage=stores[name],
                scope=scope
            )
            app.register_blueprint(
                url_prefix=self.url_prefix, blueprint=blueprints[name])

        self._oauth_blueprints = blueprints
        self._oauth_stores = stores

    @staticmethod
    def login_endpoint(app=None):
        app = app or flask.current_app
        # TODO: why isn't this just a url_for call?
        return next(
            i.rule for i in app.url_map.iter_rules()
            if i.endpoint == f"{OAuthBlueprint._oauth_name}.login_base_route")

    @staticmethod
    def session(app):
        if flask.current_app == app and not isinstance(
                flask.session, flask.sessions.NullSession):
            return flask.session
        return app.session_interface.open_session(app, flask.request)




import flask, uuid, collections, urllib.parse, time, json, datetime, qrcode, io

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from tokens import authorized
from utils import RouteLobby, CompressedUUID, data_payload, endpoint_next
from group import AccessGroup, AccessGroupRef, GroupInfo, access_stack

end_locals()

lenUUID = len(str(uuid.uuid4()))

# queries is either a str reprsenting a user UUID or a list of `users_group`s
def descendants(db, queries, selection, many=True):
    if len(queries) == 0:
        return [] if many else None
    initial, args = ("VALUES" + ",".join(("(?)",) * len(queries)), queries) \
            if type(queries) is not str else \
            ("SELECT guild FROM user_groups WHERE member=?", (queries,))
    # TODO: profile join vs subquery
    return db.many[many](
        "WITH RECURSIVE "
          "descendants(n) AS ("
            f"{initial} UNION ALL SELECT guild FROM "
            "user_groups RIGHT JOIN invitations ON via=invite, descendants "
            "WHERE inviter=descendants.n"
          ") "
        f"{selection}", args, True)

# checks if a group descends from user, returns None if not
# otherwise, returns the child and parent group of the user_group for user
def ancestors(db, initial, selection, args=(), many=True):
    return db.many[many](
        "WITH RECURSIVE "
          "ancestors(n) AS ("
            "VALUES(?) "
            "UNION ALL "
            "SELECT inviter FROM "
            "user_groups RIGHT JOIN invitations ON via=invite, ancestors "
            "WHERE guild=ancestors.n AND inviter IS NOT NULL"
          f") {selection}", (initial,) + args, True)

access_lobby = RouteLobby()

class AccessRoot:
    def __init__(self, db, redirect):
        self.registered, self.groups = [], []
        self.redirect, self.db = redirect, db
        self.bp = flask.Blueprint(
            "modular_login_access", __name__, url_prefix="/access")
        self.qual = "modular_login.modular_login_access"
        self.bp.record(lambda setup_state: self.register(setup_state.app))
        access_lobby.register_lobby(self.bp, self)

    def register(self, app):
        self.registered.append(app)
        for group in self.groups:
            group.register(app)

    def __call__(
            self, name=None, ownership_method=None, owner_id=None, sep="/",
            access_id=None, qualname=None):
        assert name is not None or access_id is not None or qualname is not None
        assert (ownership_method is None) == (owner_id is None)
        owner = owner_id and (ownership_method, owner_id)
        info = GroupInfo(self.bind, self.db, owner, sep)
        if access_id is not None or qualname is not None:
            return AccessGroupRef(self.db, sep, access_id, qualname=qualname)
        return AccessGroup(name, info)

    def bind(self, group):
        self.groups.append(group)
        for app in self.registered:
            group.register(app)

    def authorize(self):
        if not authorized():
            flask.abort(401)
        return flask.session["user"]

    def bounce(self):
        if not authorized():
            return flask.redirect(endpoint_next(self.redirect))
        return None

    def confirm(self, invite, url, **kw):
        if CompressedUUID.possible(invite):
            invite = CompressedUUID.toUUID(invite)
        implied = self.db().queryone(
            "SELECT implied FROM invitations WHERE invite=?", (invite,))
        if implied is None or implied[0] == 1:
            flask.abort(404)
        if implied[0] == 0:
            return flask.redirect(url)
        return flask.render_template(
            "confirm.html", invite=invite, url=url, **kw)

    @access_lobby.route("/accept/<invite>")
    def accept(self, invite):
        auth = self.bounce()
        if auth is not None:
            return auth
        return self.confirm(
            invite, flask.url_for(f"{self.qual}.add", invite=invite))

    @access_lobby.route("/add/<invite>")
    def add(self, invite):
        auth = self.bounce()
        if auth is not None:
            return auth
        res = self.validate(invite)
        if not isinstance(res, str):
            return res
        return flask.redirect(res)

    @access_lobby.route("/qr")
    def qr_handoff(self):
        auth = self.bounce()
        if auth is not None:
            return auth
        url = self.db().queryone(
            "SELECT redirect FROM invitations RIGHT JOIN user_groups "
            "ON via=invite WHERE redirect IS NOT NULL AND member=? "
            "ORDER BY user_groups.rowid DESC LIMIT 1", (flask.session["user"],))
        if url is None:
            flask.abort(404)
        return flask.redirect(url[0])

    @access_lobby.route("/qr/accept/<invite>")
    def qr_landing(self, invite):
        auth = self.bounce()
        if auth is not None:
            return auth
        return self.confirm(invite, qr=True, url=flask.url_for(
            f"{self.qual}.qr_add", invite=invite))

    @access_lobby.route("/qr/add/<invite>")
    def qr_add(self, invite):
        auth = self.bounce()
        if auth is not None:
            return auth
        res = self.validate(invite)
        if not isinstance(res, str):
            return res
        # TODO: add way to specify shortlink
        return flask.render_template("qr.html", url=res, handoff=flask.url_for(
            f"{self.qual}.qr_handoff", _external=True))

    @access_lobby.route("/qr/img/<invite>")
    def qr_img(self, invite):
        value = flask.url_for(
            f"{self.qual}.qr_landing", invite=invite, _external=True)
        output = io.BytesIO()
        qrcode.make(value).save(output)
        return flask.Response(output.getvalue(), mimetype='image/png')

    def preview(self, invite, info):
        short = CompressedUUID.fromUUID(invite)
        return flask.Response(flask.render_template(
            "preview.html", spots=info.invitees, url=flask.url_for(
                f"{self.qual}.accept", invite=short, _external=True),
            qr=flask.url_for(f"{self.qual}.qr_img", invite=short),
            next=flask.url_for(
                f"{self.qual}.qr_landing", invite=short, _external=True)))

    def validate(self, invite, db=None, implied=False):
        if CompressedUUID.possible(invite):
            invite = CompressedUUID.toUUID(invite)
        user = self.authorize()
        db = db or self.db().begin()
        info = db.queryone(
            "SELECT accessing, inviter, acceptance_expiration, access_limit,"
            "access_expiration, invitees, plus, depletes, dos, deauthorizes, "
            "implies, implied, redirect FROM invitations WHERE active=1 AND "
            "invite=?", (invite,), True)
        if info is None:
            db.close()
            flask.abort(410)
        if info.implied == 1 and not implied:
            db.close()
            flask.abort(400)
        now = time.time()
        # invite hasn't expired
        if info.acceptance_expiration is not None and \
                info.acceptance_expiration < now or \
                info.access_expiration is not None and \
                info.access_expiration < now or \
                info.access_limit is not None and \
                info.access_limit < now:
            db.close()
            self.db().execute(
                "UPDATE invitations SET active=0 WHERE invite=?", (invite,))
            flask.abort(401)
        user_group = info.inviter and db.queryone(
            "SELECT member FROM user_groups WHERE guild=?",
            (info.inviter,), True)
        if user_group and user_group.member == user:
            db.close()
            return self.preview(invite, info)
        # can't accept the same invite twice
        previously = db.queryone(
            "SELECT guild FROM user_groups WHERE via=? AND member=?",
            (invite, user), True)
        if previously:
            db.close()
            flask.abort(400, description=f"already accepted: {previously}")
        # no user_group loops
        # lower depletions; invites' invitees is not None if depletes
        chain = [] if info.inviter is None else ancestors(
            db, info.inviter, "SELECT guild, member, spots, invite, invitees, "
            "depletes FROM user_groups LEFT JOIN invitations on via=invite "
            "WHERE guild IN ancestors")
        if any(i.member == user for i in chain):
            db.close()
            flask.abort(412)
        if info.invitees == 0:
            db.close()
            flask.abort(404)
        invite_update, user_update = [], []
        if info.invitees is not None:
            invite_update.append((info.invitees - 1, invite))
        for i, level in enumerate(chain):
            if level.spots == 0 or level.invitees == 0:
                db.close()
                flask.abort(404)
            if level.spots is not None:
                user_update.append((level.spots - 1, level.guild))
            if level.invitees is not None:
                invite_update.append((level.invitees - 1, level.invite))
            if not level.depletes:
                break
        db.executemany(
            "UPDATE user_groups SET spots=? WHERE guild=?", user_update)
        db.executemany(
            "UPDATE invitations SET invitees=? WHERE invite=?", invite_update)
        creating = str(uuid.uuid4())
        until = info.access_expiration
        if until is not None and until < 0:
            until = min(now + until, info.access_limit)
        db.execute(
            "INSERT INTO user_groups(guild, via, member, "
            "access_group, until, spots) VALUES (?, ?, ?, ?, ?, ?)", (
                creating, invite, user, info.accessing, until,
                info.plus))
        dos = info.dos and (info.dos - 1)
        if info.implies is not None:
            return self.validate(info.implies, db, True)
        db.commit().close()
        return info.redirect

    @staticmethod
    # selecting columns needed to know what invites selected can create
    def group_query(db, member=None, access_groups=()):
        assert member or access_group
        member_query = ((), ()) if not member else (("member=?",), (member,))
        access_groups_query = ((), ()) if not access_groups else ((
            "user_groups.access_group IN (" +
            ", ".join(("?",) * len(access_groups)) + ")",), access_groups)
        return db.queryall(
            "SELECT access_group, guild, via, member, " +
            "until, spots, depletes, dos, " +
            "CASE WHEN deauthorizes IS NULL THEN 0 ELSE deauthorizes END AS " +
            "deauthorizes FROM user_groups " +
            "LEFT JOIN invitations ON invite=via WHERE " +
            "user_groups.active=1 AND " +
            "(until IS NULL or until>unixepoch()) AND " +
            " AND ".join(member_query[0] + access_groups_query[0]),
            member_query[1] + access_groups_query[1], True)

    access_info = collections.namedtuple("AccessInfo", (
        "access_group", "guild", "via", "member", "until",
        "spots", "depletes", "dos", "deauthorizes", "depletion_bound",
        "implied_groups"))

    @staticmethod
    def depletion_bound(db, initial):
        opt = lambda *a: None if all(i is None for i in a) else min(
            i for i in a if i is not None)
        minimum, chain = None, ancestors(
            db, initial, "SELECT spots, invitees, depletes "
            "FROM user_groups RIGHT JOIN invitations on via=invite "
            "WHERE guild IN ancestors")
        for level in chain:
            minimum = opt(minimum, level.spots)
            if level.depletes:
                minimum = opt(minimum, level.invitees)
            else:
                break
        return minimum

    # returns access_info for all groups user is in
    def user_groups(self, user=None, groups=None, db=None):
        user = user or flask.session["user"]
        db, close = db or self.db().begin(), db is None
        info = self.group_query(db, user, () if groups is None else groups)
        results = []
        for option in info:
            subgroups = db.queryall(
                "WITH RECURSIVE "
                  "subsets(n) AS ("
                    "VALUES(?) "
                    "UNION ALL "
                    "SELECT access_id FROM access_groups, subsets "
                    "WHERE parent_group=subsets.n"
                  ") "
                "SELECT access_id, group_name FROM access_groups "
                "WHERE access_id IN subsets", (option.access_group,))
            results.append(option + (
                self.depletion_bound(db, option.guild), subgroups))
        if close:
            db.close()
        return [self.access_info(*option) for option in results]

    # returns all members of a given list of groups
    # shouldn't necessarily be viewable
    def group_access(self, groups, db=None):
        db, close = db or self.db().begin(), db is None
        stack = {}
        for group in groups:
            stack[group] = access_stack(
                db, group, "SELECT uuid, group_name FROM access_groups WHERE "
                "uuid IN supersets")
        uniq = set(sum([i[0] for i in stack.values()], []))
        info = self.group_query(db, access_groups=uniq)
        results = [self.access_info(
            *option, self.depletion_bound(db, option.guild),
            stack[option.access_group]) for option in info]
        if close:
            db.close()
        return results

    @access_lobby.template_json(
        "/invite/<group>", "invite.html", methods=["GET", "POST"])
    def single_group_invite(self, group):
        return self.invite(group)

    @access_lobby.template_json("/invite", "invite.html")
    @access_lobby.route("/view/invite", methods=["POST"])
    def invite(self, group=None):
        user = self.authorize()
        if flask.request.method == "GET":
            memberships = self.user_groups(user, group and (group,))
            invitable = [
                i for i in memberships if
                (i.depletion_bound is None or i.depletion_bound > 0) and
                (i.dos is None or i.dos > 1)]
            return {"groups": invitable}
        else:
            return self.parse_invite(flask.request.form)

    creation_args = {
        "invitees": {None, int},
        "redirect": str,
        "confirm": bool,
        "invitations": [{
            "accessing": str,
            "acceptance_expiration": {None, int},
            "access_expiration": {None, int},
            "plus": {None, int},
            "inviter": str,
            "depletes": bool,
            "dos": {None, int},
            "deauthorizes": {0, 1, 2},
        }]}

    @access_lobby.route("/allow", methods=["POST"])
    def allow(self):
        return json.dumps(self.create(flask.request.json))

    def parse_invite(self, form):
        if not form.get("redirect"):
            flask.abort(400, description="missing redirect")
        payload = {
            "redirect": form.get("redirect"), "confirm": "confirm" in form,
            "invitees": form.get("invitees") or None, "invitations": []}
        try:
            tz = datetime.timezone(
                -datetime.timedelta(minutes=int(form.get("tz", 0))))
        except ValueError:
            flask.abort(400, description="invalid tz")
        if payload["invitees"] is not None:
            try:
                payload["invitees"] = int(payload["invitees"])
            except ValueError:
                flask.abort(400, description="invalid invitees")
        payload["invitations"] = [
            {"accessing": i, **{
                k[:-lenUUID - 1]: form[k] for k in form.keys()
                if k.endswith(i) and k != i}}
            for i in form.keys() if len(i) == lenUUID and i[-22] == "4"]
        if len(payload["invitations"]) == 0:
            flask.abort(400, description="no invite groups")
        for group in payload["invitations"]:
            for dated in ("access_expiration", "acceptance_expiration"):
                value = group.get(dated)
                if value:
                    group[dated] = int(datetime.datetime.fromisoformat(
                        value).replace(tzinfo=tz).timestamp())
                else:
                    group[dated] = None
            relative = group.pop("access-num", None)
            if group.pop(
                    "expiration-type", "").startswith("relative") and relative:
                group["access_expiration"] = -int(float(relative) * 86400)
            group['depletes'] = 'depletes' in group
            for num in ("plus", "dos", "deauthorizes"):
                try:
                    group[num] = None if group.get(num) in (None, '') \
                        else json.loads(group[num])
                except ValueError:
                    flask.abort(400, description=f"invalid {num}")
        return flask.redirect(flask.url_for(
            f"{self.qual}.add",
            invite=self.create(payload)["long"]))

    def create(self, payload):
        inserting = (
            "accessing", "acceptance_expiration", "access_expiration", "plus",
            "inviter", "depletes", "dos", "deauthorizes")
        payload = data_payload(payload, self.creation_args, True)
        user = self.authorize()
        db = self.db().begin()
        if len(payload.invitations) == 0:
            db.close()
            flask.abort(400, description="no invites")
        if not payload.redirect:
            db.close()
            flask.abort(400, description="bad redirect")
        values, first = [], (i == 0 for i in range(len(payload.invitations)))
        current_uuid, next_uuid = None, None
        now = time.time()
        for invite, last in reversed(tuple(zip(payload.invitations, first))):
            # user accepted via
            # user has access to group through via (limitations.active = 1)
            users_group = access_stack(db, invite.accessing,
                "SELECT via, until, spots FROM user_groups WHERE "
                "guild=? AND member=? AND active=1 AND "
                "(until IS NULL or until>unixepoch()) AND "
                "access_group IN supersets", (invite.inviter, user), False)
            if users_group is None:
                db.close()
                flask.abort(401, description="invalid source")
            if users_group.via is None:
                limits = collections.namedtuple("limits", (
                    "depletes", "dos", "deauthorizes", "plus"))(
                        False, None, 0, None)
            else:
                limits = db.queryone(
                    "SELECT depletes, dos, deauthorizes, plus FROM invitations "
                    "WHERE invite=?", (users_group.via,), True)
            # acceptance expiration and access expiration are before until
            #     (negative values for access expiration are after acceptance)
            #     (limited by access_limit)
            # TODO: redundant logic in /add
            if users_group.until is not None and (
                    invite.acceptance_expiration is None or
                    invite.acceptance_expiration > users_group.until or
                    invite.access_expiration is None or
                    invite.access_expiration > users_group.until):
                db.close()
                flask.abort(400, description="unauthorized timing")
            if invite.acceptance_expiration is not None and \
                    invite.acceptance_expiration < now or \
                    invite.access_expiration is not None and \
                    0 < invite.access_expiration < now:
                db.close()
                flask.abort(400, description="invalid timing")
            if invite.acceptance_expiration is not None and \
                    invite.access_expiration is not None and \
                    0 < invite.access_expiration and \
                    invite.access_expiration < invite.acceptance_expiration:
                db.close()
                flask.abort(400, description="invalid timing")
            # invitees is less than or equal to depletion bound
            bound = self.depletion_bound(db, invite.inviter)
            if bound is not None and (
                    payload.invitees is None or payload.invitees > bound):
                db.close()
                flask.abort(400, description="too many invitees")
            if users_group.spots is not None and (
                    invite.plus is None or invite.plus > users_group.spots):
                db.close()
                flask.abort(400, description="too many plus ones")
            # 0 < dos < limits.dos
            if limits.dos is not None if invite.dos is None else (
                    0 > invite.dos and (
                        limits.dos is None or invite.dos >= limits.dos)):
                db.close()
                flask.abort(400, description="invalid degrees of separation")
            # invite deauthorizes <= limitations.deauthorizes
            if invite.deauthorizes > limits.deauthorizes:
                db.close()
                flask.abort(401, description="can't deauthorize")
            # depletes >= limits.depletes
            if not invite.depletes and limits.depletes:
                db.close()
                flask.abort(401, description="must deplete")
            current_uuid, next_uuid = str(uuid.uuid4()), current_uuid
            implied = (-1 if payload.confirm else 0) if last else 1
            redirect = payload.redirect if next_uuid is None else None
            # TODO: what about "try for 3 days" spreadable invites
            values.append(tuple(getattr(invite, i) for i in inserting) + (
                payload.invitees, users_group.until, implied, redirect,
                current_uuid, next_uuid))
        db.executemany(
            "INSERT INTO invitations(" + ", ".join(inserting) +
            ", invitees, access_limit, implied, redirect, invite, implies) " +
            "VALUES (" + ", ".join(("?",) * (len(inserting) + 6)) + ")", values)
        db.commit().close()
        return {
            "long": current_uuid,
            "short": CompressedUUID.fromUUID(current_uuid)}

    removal_args = [str] # user_group UUIDs

    @access_lobby.route("/revoke", methods=["POST"])
    def revoke(self):
        self.kick(flask.request.json)
        return flask.request.data

    def stack_auth(self, db, user, access_group):
        return access_stack(
                db, access_group, "SELECT MAX"
                "(CASE WHEN deauthorizes IS NULL THEN 0 ELSE deauthorizes END) "
                "AS deauthorizes FROM invitations "
                "RIGHT JOIN user_groups ON via=invite "
                "WHERE user_groups.active=1 AND member=? AND "
                "(until IS NULL or until>unixepoch()) AND "
                "access_group IN supersets", (user,), False).deauthorizes

    def kick(self, payload):
        user = self.authorize()
        payload = data_payload(payload, self.removal_args, True)
        db = self.db().begin()
        access_groups = dict(db.queryall(
            "SELECT guild, access_group FROM user_groups WHERE guild IN (" +
            ", ".join(("?",) * len(payload)) + ")",
            [revoking for revoking in payload]))
        for revoking in payload:
            access_group = access_groups.get(revoking)
            # also ensures privledges query is not empty
            if access_group is None:
                flask.abort(400)
            privledges = self.stack_auth(db, user, access_group)
            if privledges == 0:
                db.close()
                flask.abort(401)
            if privledges == 1:
                # walk stack to check if user has any ancestor groups from
                # user_group with privledges to deauthorize
                # though under the current setup, there should only be one
                # instance of user along any path from root
                if not ancestors(
                        db, revoking, "SELECT guild FROM "
                        "user_groups RIGHT JOIN invitations ON "
                        "via=invite WHERE deauthorizes=1 AND "
                        "member=? AND guild IN ancestors", (user,)):
                    db.close()
                    flask.abort(401)
            # no need to check privledges for deauthorizes == 2
        db.executemany(
            "UPDATE user_groups SET active=0 WHERE guild=?",
            [(guild,) for guild in payload])
        db.commit().close()

    deauth_info = collections.namedtuple("Deauthable", (
        "member", "display_name", "user_group", "access_group", "group_name"))

    # returns member, user_group, access_group, group_name
    def deauthable(self, user, db=None):
        db = db or self.db().begin()
        groups = self.user_groups(user, db=db)
        permissions = [[], [], []]
        for group in groups:
            permissions[group.deauthorizes].append(group)
        # permissions[1] access_group, group_name (descendants all in implied)
        # member, access_group, uuid
        childrens_groups = descendants(
            db, [group.guild for group in permissions[1]],
            "SELECT via, guild, member, access_group FROM user_groups "
            "WHERE guild IN descendants")
        # permissions[2] access_group, group_name
        implied = set() if len(permissions[2]) == 0 else set(
            i[0] for group in permissions[2] for i in group.implied_groups)
        # member, uuid, access_group
        subgroupers = [] if len(implied) == 0 else db.queryall(
            "SELECT member, guild, access_group FROM user_groups " +
            "WHERE user_groups.active=1 AND access_group IN (" +
            ", ".join(("?",) * len(implied)) + ")", list(implied), True)
        members = set(
            share.member for share in
            subgroupers + childrens_groups + permissions[0])
        display_names = dict(db.queryall(
            "SELECT uuid, display_name FROM auths WHERE uuid IN (" +
            ", ".join(("?",) * len(members)) + ")", tuple(members)))
        # python joins because of recursive queries
        match = [
            dict(sum([share.implied_groups for share in level], []))
            for level in permissions[1:]]
        results = [[
            self.deauth_info(
                share.member, display_names[share.member], share.guild,
                share.access_group, share.implied_groups[0][1])
            for share in permissions[0]]]
        for group_names, user_group in zip(match, (
                childrens_groups, subgroupers)):
            results.append([
                self.deauth_info(
                    share.member, display_names[share.member], share.guild,
                    share.access_group, group_names[share.access_group])
                for share in user_group])
        return results

    @access_lobby.template_json("/remove", "remove.html")
    @access_lobby.route("/view/remove", methods=["POST"])
    def remove(self):
        if flask.request.method == "GET":
            user = self.authorize()
            return {"removable": self.deauthable(user)}
        else:
            removing = list(flask.request.form.keys())
            self.kick(removing)
            return json.dumps(removing)

    invite_cols = 15
    invite_info = collections.namedtuple("InviteInfo", (
        "invite", "accessing", "inviter", "acceptance_expiration",
        "access_expiration", "access_limit", "invitees", "plus", "depletes",
        "dos", "deauthorizes", "implies", "implied", "redirect", "active",
        "group_name", "display_name"))

    # returns member, user_group, access_group, group_name
    def hosting(self, user, db=None):
        db = db or self.db().begin()
        groups = self.user_groups(user, db=db)
        permissions = [[], [], []]
        for group in groups:
            permissions[group.deauthorizes].append(group)
        selection = ", ".join(self.invite_info._fields[:self.invite_cols])
        childrens_invites = descendants(
            db, [group.guild for group in permissions[1]],
            f"SELECT {selection} FROM invitations WHERE inviter IN descendants "
            "AND active=1")
        flat = tuple(set(sum([
            group.implied_groups[0] for group in permissions[2]], ())))
        group_invites = [] if len(flat) == 0 else db.queryall(
            f"SELECT {selection} FROM invitations WHERE accessing IN (" +
            ", ".join(("?",) * len(flat)) + ") AND active=1", flat, True)
        personal_invites = [] if len(permissions[0]) == 0 else db.queryall(
            f"SELECT {selection} FROM invitations WHERE inviter IN (" +
            ", ".join(("?",) * len(permissions[0])) + ") AND active=1", [
                group.guild for group in permissions[0]], True)
        levels = [personal_invites, childrens_invites, group_invites]
        directory = dict(sum([group.implied_groups for group in groups], []))
        user_groups = set(filter(None, (
            group.inviter for group in sum(levels, []))))
        usernames = {None: None, **dict(db.queryall(
            "SELECT guild, display_name FROM auths "
            "RIGHT JOIN user_groups ON member=auths.uuid WHERE guild IN (" +
            ", ".join(("?",) * len(user_groups)) + ")", tuple(user_groups)))}
        return [[
            self.invite_info(
                *group, directory[group.accessing], usernames[group.inviter])
            for group in level] for level in levels]

    renege_args = [str]

    def uninvite(self, payload):
        user = self.authorize()
        payload = data_payload(payload, self.renege_args, True)
        db = self.db().begin()
        access_groups = {i[0]: i[1:] for i in db.queryall(
            "SELECT invite, accessing, inviter FROM invitations " +
            "WHERE invite IN (" + ", ".join(("?",) * len(payload)) + ")",
            [reneging for reneging in payload])}
        for reneging in payload:
            access_group = access_groups.get(reneging, [None])
            # also ensures privledges query is not empty
            if access_group[0] is None:
                flask.abort(400)
            privledges = self.stack_auth(db, user, access_group[0])
            if access_group[1] is None and privledges != 2:
                flask.abort(401)
            if privledges == 0 and (user,) != db.queryone(
                    "SELECT member FROM user_groups WHERE guild=?",
                    (access_group[1],)):
                db.close()
                flask.abort(401)
            if privledges == 1 and not ancestors(
                    db, access_group[1], "SELECT guild FROM "
                    "user_groups RIGHT JOIN invitations ON "
                    "via=invite WHERE deauthorizes=1 AND "
                    "member=? AND guild IN ancestors", (user,)):
                db.close()
                flask.abort(401)
        db.executemany(
            "UPDATE invitations SET active=0 WHERE invite=?",
            [(invite,) for invite in payload])
        db.commit().close()

    @access_lobby.template_json("/renege", "renege.html")
    @access_lobby.route("/view/renege", methods=["POST"])
    def renege(self):
        if flask.request.method == "GET":
            return {"hosts": self.hosting(self.authorize())}
        else:
            removing = list(flask.request.form.keys())
            self.uninvite(removing)
            return json.dumps(removing)

    @access_lobby.route("/cancel", methods=["POST"])
    def cancel(self):
        self.uninvite(flask.request.json)
        return flask.request.data




import collections, uuid, json

import sys, os.path; end_locals, start_locals = lambda: sys.path.pop(0), (
    lambda x: x() or x)(lambda: sys.path.insert(0, os.path.dirname(__file__)))

from utils import InfoShell

end_locals()

GroupInfo = collections.namedtuple("GroupInfo", ("bind", "db", "owner", "sep"))

# returns stack of access groups going upwards from init
def access_stack(db, init, query, args=(), many=True):
    init = (init if " " in init or "(" in init else (init,)) \
        if type(init) is str else init
    init, select = ((), init) if type(init) is str else \
        (init, f"VALUES{','.join(('(?)',) * len(init))}")
    return db.many[many](
        "WITH RECURSIVE "
          "supersets(n) AS ("
            f"{select} "
            "UNION ALL "
            "SELECT parent_group FROM access_groups, supersets "
            "WHERE access_id=supersets.n"
          f") {query}", init + args, True)

# group can be a root last stack
def ismember(db, user, group, args=()):
    return access_stack(
        db, group, "SELECT guild, access_group FROM user_groups WHERE "
        "access_group IN supersets AND member=? AND active=1 AND "
        "(until IS NULL or until>unixepoch())", args + (user,), False)

class AccessGroup(InfoShell):
    def __init__(self, name, info, stack=None):
        self.info = info
        assert info.sep not in name
        self.name, self.uuid = name, None
        self.stack, self.root = (stack or []) + [self], not bool(stack)
        self.info.bind(self)

    def group(self, name):
        return __class__(name, self.info, self.stack)

    @property
    def qualname(self):
        return self.info.sep.join(i.name for i in self.stack)

    def __repr__(self):
        return self.qualname

    def register(self, app):
        db = self.db(app).begin()
        self.ensure(db)
        db.close()

    def ensure(self, db):
        if self.uuid is not None:
            return
        parent = None if self.root else self.stack[-2].uuid
        access = db.queryone(
            "SELECT parent_group, access_id FROM access_groups "
            "WHERE group_name=?", (self.qualname,), True)
        uniq = str(uuid.uuid4()) if access is None else access.access_id
        if access is None or access.parent_group != parent:
            # update if the access_group structure has changed
            db.execute(
                "INSERT INTO access_groups(group_name, parent_group, access_id)"
                "VALUES (?, ?, ?) ON CONFLICT(group_name) DO UPDATE SET "
                "parent_group=excluded.parent_group",
                (self.qualname, parent, uniq))
            self.uuid = uniq
        else:
            self.uuid = access.access_id

        if self.info.owner is not None and self.root:
            owner = db.queryone(
                "SELECT uuid FROM auths WHERE method=? AND platform_id=?",
                self.info.owner)
            if owner is None:
                owner = str(uuid.uuid4())
                db.execute(
                    "INSERT INTO auths(method, platform_id, uuid) "
                    "VALUES (?, ?, ?)", self.info.owner + (owner,))
                changed=True
            else:
                owner = owner[0]
                changed = db.queryone(
                    "SELECT 1 FROM user_groups LEFT JOIN invitations "
                    "ON via=invite WHERE until IS NULL AND spots IS NULL AND "
                    "user_groups.active=1 AND member=? AND access_group=? AND "
                    "deauthorizes=2", (owner, self.uuid)) is None
            if changed:
                invite = str(uuid.uuid4())
                db.execute(
                    "INSERT INTO invitations"
                    "(invite, accessing, inviter, depletes, deauthorizes) "
                    "VALUES (?, ?, NULL, FALSE, 2)", (invite, self.uuid))
                db.execute(
                    "INSERT INTO user_groups"
                    "(guild, via, member, access_group, spots) "
                    "VALUES (?, ?, ?, ?, NULL)",
                    (str(uuid.uuid4()), invite, owner, self.uuid))
        db.commit()

    def shallow(self, app, user):
        db = self.db(app).begin()
        res = ismember(db, user, [self.uuid])
        db.close()
        return res

    def db(self, app=None):
        return self.info.db() if app is None else self.info.db(app).ctx

    # TODO: strict ordering (see Google Zanzibar) using read/write decorators?
    def vet(self, user, app=None):
        db = self.db(app).begin()
        res = ismember(db, user, tuple(reversed([i.uuid for i in self.stack])))
        db.close()
        return res

    def __contains__(self, user):
        return bool(self.vet(user))

    def __truediv__(self, other):
        assert isinstance(other, str)
        return AccessGroupRef(self.info.db, self.info.sep, None, self, other)

    def add_user(self, user):
        guild = str(uuid.uuid4())
        self.db().execute(
            "INSERT INTO user_groups(guild, member, access_group) "
            "VALUES (?, ?, ?)", (guild, user, self.uuid))
        return guild

    def remove_access(self, guild):
        return self.db().execute(
            "UPDATE user_groups SET active=0 WHERE guild=?", (self.uuid,))

class AccessGroupRef(AccessGroup):
    def __init__(
            self, db, sep='/', access_id=None, source=None, *names,
            qualname=None, owner=None):
        self.db, self.sep, self._name = lambda *a, **kw: db(), sep, qualname
        self.info = GroupInfo(None, self.db, owner, sep)
        assert access_id and not names or names or qualname and not source
        self._uuid, self.source, self.names = access_id, source, names
        if self._name is not None:
            prefix = self._name.split(sep)
            if self.names:
                self._name = None
            self.names = tuple(prefix) + self.names

    def __contains__(self, user):
        if self.source is not None and self.names or self._name:
            return bool(ismember(
                self.db(), user,
                "SELECT access_id FROM access_groups WHERE group_name=?",
                (self.qualname,)))
        elif self._stack is None and self._uuid is not None:
            return bool(ismember(self.db(), user, self.uuid))
        else:
            return super().__contains__(user)

    @property
    def qualname(self):
        if self._name is None:
            if self.names:
                self._name = self.sep.join(self.names)
                if self.source is not None:
                    self._name = self.source.qualname + self.sep + self._name
            else:
                self._name = self.db().queryone(
                    "SELECT group_name FROM access_groups WHERE uuid=?",
                    (self.uuid,))
                assert self._name is not None
                self._name = self._name[0]
        return self._name

    @property
    def uuid(self):
        if self._uuid is None:
            self._uuid = self.db().queryone(
                "SELECT access_id FROM access_groups WHERE group_name=?",
                (self.qualname,))
            if self._uuid is not None:
                self._uuid = self._uuid[0]
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value
        if self._stack is not None:
            self._stack[0] = self.access_ref()

    access_ref = lambda self: type("AccessRef", (), {"uuid": self.uuid})()
    _stack = None
    @property
    def stack(self):
        if self._stack is None:
            if len(self.names) == 1 and self.source is not None:
                self._stack = [self.access_ref()] + self.source.stack
            elif self._uuid is None:
                self._stack = access_stack(
                    self.db(),
                    "SELECT access_id FROM access_groups WHERE group_name=?",
                    "SELECT n AS uuid FROM supersets", (self.qualname,))
                assert len(self._stack) > 0
                self._uuid = self._stack[0].uuid
            else:
                self._stack = access_stack(
                    self.db(), self.uuid, "SELECT n AS uuid FROM supersets")
        return self._stack

    _root = None;
    @property
    def root(self):
        if self._root is None:
            self._root = self.source is None and len(self.names) == 1
        return self._root

    def __truediv__(self, other):
        assert isinstance(other, str)
        if self.source is None or self._stack is not None:
            return __class__(self.db, self.sep, None, self, other)
        return __class__(
            self.db, self.sep, None, self.source, *(self.names + (other,)))

    @classmethod
    def reconstruct(cls, db, rpn, sep='/', owner=None):
        def f(el):
            return cls(db, sep, qualname=el, owner=owner)
        if isinstance(rpn, str):
            if '"' in rpn:
                rpn = json.loads(rpn)
            else:
                return f(rpn)
        assert isinstance(rpn, list)

        def inner(el):
            return [inner(i) if type(i) is list else f(i) for i in el[1:]]
        args = inner(rpn)
        return getattr(args[0], rpn[0])(*args[1:])

    def ensure(self, db=None):
        db, close = db or self.db().begin(), db is None
        if len(self.names) > 1:
            self.source = __class__(
                self.db, self.sep, None, self.source, *self.names[:-1],
                owner=self.info.owner)
            self.names = (self.names[-1],)
        if self.source is not None:
            self.source.ensure(db)
        super().ensure(db)
        if close:
            db.close()

